#
# jomiel-comm
#
# Copyright
#  2019 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""


def exit_error():
    """Wraps the sys.exit call, exits with an error code (1)."""
    from sys import exit as _exit

    _exit(1)


def load_cert_file(path, logger=None):
    """Loads a certificate file from the given location.

    Notes:

        - Exits with an error if zmq.auth.load_certificate raised an
          IOError (e.g. file not found)

    """
    from zmq.auth import load_certificate
    from os.path import abspath

    keyfile = abspath(path)

    try:
        return load_certificate(keyfile)
    except OSError as message:
        if logger:
            logger.error(message)
        exit_error()


def setup(socket, curve, logger=None):
    """Load the CURVE certificates required for a secure client connection.

    Args:
        socket (obj): the zmq socket to use

        curve (obj): the named tuple containing the curve options

        logger (obj): the logger instance to use (or None)

    Raises:
        ValueError if any of the required kwarg values were unacceptable

    Returns:
        Nothing

    """
    (pubkey, seckey) = load_cert_file(curve.client_key_file, logger)

    socket.curve_secretkey = seckey
    socket.curve_publickey = pubkey

    (pubkey, _) = load_cert_file(curve.server_public_key_file, logger)
    socket.curve_serverkey = pubkey


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-kore
#
# Copyright
#  2019-2020 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""
from abc import ABCMeta
from abc import abstractmethod
from importlib import import_module
from sys import stdout

try:  # py37+
    from importlib.resources import path as resources_path
except ModuleNotFoundError:
    from importlib_resources import path as resources_path


class App(metaclass=ABCMeta):
    """A simple core class that wraps all-things-necessary to create
    command line interface application with very little effort."""

    __slots__ = [
        "_no_default_config_files",
        "_no_version_long_option",
        "_no_print_config_option",
        "_no_config_file_option",
        "_no_logger_options",
        "_config_module",
        "_logger_files",
        "_package_additional_search_paths",
        "_package_data_dir",
        "_package_name",
        "_version",
    ]

    def __init__(self, **kwargs):
        """Initializes the object.

        Args:
            **kwargs: arbitrary keyword args

        Supported arbitrary keyword args (kwargs):

            package_name (str): The package name to be used for
                package resource queries and building the XDG XDG
                configuration file search paths.

                - The value will be used to determine the different XDG
                  configuration file paths

                - If None, the package resources queries will be skipped

                - If None, XDG configuration file path searches will be
                  skipped

                The value is typically set to __name__ by the caller.

            package_data_dir (str): The package data dir. This is
                package_name + package_data_dir. Be sure to set
                package_name, also.

            package_additional_search_paths (list): The _additional_
                package search paths to be added to the sys.path.

            config_module (str): Module path to the "configuration
                module" used throughout the application runtime
                life-cycle to access global _cached_ values.

            version (str): The program version string, or None.

            no_print_config_option (bool): If True, disable -D and -E
                options.

            no_default_config_files (bool): If True, skip the XDG paths
                in the configuration file search.

            no_version_long_option (bool): If True, disable
                --version-long option, altogether.

            no_config_file_option (bool): If True, disable the
                --config-file option.

            no_logger_options (bool): If True, disable all of the logger
                options (e.g. --logger-file, --logger-ident).

        """
        self._no_version_long_option = kwargs.get(
            "no_version_long_option", False,
        )
        self._no_print_config_option = kwargs.get(
            "no_print_config_option", False,
        )
        self._no_logger_options = kwargs.get("no_logger_options", False)
        self._config_module = kwargs.get("config_module")

        self._package_additional_search_paths = kwargs.get(
            "package_additional_search_paths",
        )
        self._package_data_dir = kwargs.get("package_data_dir")
        self._package_name = kwargs.get("package_name")

        self._no_default_config_files = kwargs.get(
            "no_default_config_files", False,
        )

        self._no_config_file_option = kwargs.get(
            "no_config_file_option", False,
        )

        def add_package_search_paths():
            """Append the additional package search paths to sys.path."""
            if self._package_additional_search_paths:
                from sys import path as sys_path

                for mod_path in self._package_additional_search_paths:
                    with resources_path(mod_path, "") as path:
                        sys_path.append(str(path))

        def determine_xdg_paths():
            """Return the XDG paths to configuration files."""

            pkg_name = self._package_name

            if not pkg_name or self._no_default_config_files:
                return ([], [])

            config_files = [
                "/etc/xdg/{0}/{0}.yaml".format(pkg_name),
                "~/.config/{0}/{0}.yaml".format(pkg_name),
                f"./{pkg_name}.yaml",
            ]

            logger_files = [
                f"/etc/xdg/{pkg_name}/logger.yaml",
                f"~/.config/{pkg_name}/logger.yaml",
                "./logger.yaml",
            ]

            if pkg_name:
                path_prefix = pkg_name

                if self._package_data_dir:
                    path_prefix = self._package_data_dir

                config_path = f"{path_prefix}.config.logger"
                config_fname = f"{pkg_name}.yaml"

                try:
                    with resources_path(config_path, config_fname) as p:
                        logger_files.insert(0, p)
                except FileNotFoundError:
                    pass

            return (config_files, logger_files)

        add_package_search_paths()

        (config_files, self._logger_files) = determine_xdg_paths()
        self._version = kwargs.get("version", "?")

        from configargparse import get_parser

        parser = get_parser(
            default_config_files=config_files,
            add_config_file_help=False,
        )

        parser.add(
            "--version",
            action="version",
            version="%(prog)s version " + self._version,
        )

        if not self._no_version_long_option:
            parser.add(
                "-v",
                "--version-long",
                help="""show version information about program's
                        environment and exit""",
                action="store_true",
            )

        if not self._no_config_file_option:
            parser.add(
                "--config-file",
                help="Read configuration from the specified file",
                is_config_file=True,
                metavar="FILE",
            )

        if not self._no_print_config_option:
            parser.add(
                "-D",
                "--print-config",
                help="Show the configuration values and exit",
                action="store_true",
            )

            parser.add(
                "-E",
                "--report-config",
                help="Report keys, values and where they were set",
                action="store_true",
            )

        def logger_group():
            """Add the logger option group."""
            grp = parser.add_argument_group("logger")

            grp.add(
                "--logger-config",
                help="Logger configuration file to read",
                metavar="FILE",
            )

            grp.add(
                "-L",
                "--logger-idents",
                help="Print logger identities and exit",
                action="store_true",
            )

            grp.add(
                "--logger-idents-verbose",
                help="Print logger identities in detail, "
                "use together with --logger-idents",
                action="store_true",
            )

            grp.add(
                "-l",
                "--logger-ident",
                help="Use the logger identity",
                metavar="IDENT",
                default="default",
            )

        if not self._no_logger_options:
            logger_group()

    @abstractmethod
    def run(self):
        """[Override] Runs the program."""

    def parse_opts(self, parser):
        """Parses the options.

        Notes:
            - Handles the -D and -E options gracefully

            - Applications subclassing this class, do not need to
              implement the support

            - Instead, the these are either disabled by the subclass or
              handled automagically by the superclass

        Args:
            parser (obj): configargparse parser instance

        """

        def handle_print_config():
            """Handle the -D and -E options."""

            if self._no_print_config_option:
                return

            def print_config_values(opts):
                """Prints the configuration values to stdout and
                terminate the program.

                Args:
                    opts (dict): configargparse returned options

                """
                data = {
                    "configuration": opts.__dict__,
                }
                dump_as_yaml(data)

            def print_report_config(parser):
                """Prints the configuration sources to stdout and
                terminates the program.

                Args:
                    parser (obj): configargparse parser instance

                """
                parser.print_values()
                exit_normal()

            if opts.print_config:
                print_config_values(opts)

            elif opts.report_config:
                print_report_config(parser)

        def handle_version_long():
            """Handle --version-long"""

            if self._no_version_long_option or not opts.version_long:
                return

            def version_long():
                """Return string to be printed with --version-long"""

                def package_versions():
                    """Return the package versions."""
                    from .version import package_version

                    required_packages = self.version_long_packages()
                    found_packages = []

                    for package_name in sorted(required_packages):
                        package_version(
                            package_name, found_packages,
                        )

                    return [
                        {key: value} for key, value in found_packages
                    ]

                from sys import version as py_version

                return {
                    "version": self._version,
                    "python": {
                        "version": py_version.replace("\n", ""),
                        "packages": package_versions(),
                    },
                }

            yaml = version_long()
            dump_as_yaml(yaml)

        def setup_global_config():
            """Sets up the global config module with the parsed options."""

            if not self._config_module:
                return

            mod = import_module(self._config_module)

            if not self._no_logger_options:
                if opts.logger_config:
                    self._logger_files.insert(0, opts.logger_config)
                mod.logger_paths = self._logger_files

            mod.opts = opts

        opts = parser.parse()

        handle_print_config()
        handle_version_long()

        setup_global_config()

        return opts

    def version_long_packages(self):
        """[Override] Returns the required packages that we wish to check
        with --version-long and append to the output.

        Note:
            - Override this in your applications

        Returns:
            list: of package names (empty)

            Example:
                return ['requests', 'yaml', 'pyzmq']

        """
        return []


def subprocess_open(args):
    """Execute subprocess by using Popen."""
    from subprocess import Popen, PIPE

    chld = Popen(args, stdout=PIPE)
    data = chld.communicate()[0]
    return (chld.returncode, data.rstrip().decode("utf8"))


def exit_with(code):
    """Calls sys.exit with the exit status code."""
    from sys import exit as _exit

    _exit(code)


def exit_error():
    """Wraps the sys.exit call, exits with an error status code (1)."""
    exit_with(1)


def exit_normal():
    """Wraps the sys.exit call, exits with an OK status code (0)."""
    exit_with(0)


def round_trip_dump_yaml(data, stream=None):
    """Dump YAML.

    Args:
        data (dict): to be dumped as YAML
        stream (obj): to dump the YAML to

    """
    from ruamel.yaml import YAML, round_trip_dump

    yaml = YAML(typ="safe")
    yaml.default_flow_style = False
    round_trip_dump(data, stream)


def dump_as_yaml(yaml):
    """Dump the given data as YAML to stdout and exit process.

    Args:
        yaml (dict): data to be dumped

    """
    stdout.write("---\n")
    round_trip_dump_yaml(yaml, stdout)
    exit_normal()


def dump_logger_identities(loggers, detailed=False):
    """Dump the found logger identities to the stdout and exit.

    Args:
        loggers (dict): from the parsed .yaml to be dumped
        detailed (bool): be detailed

    """
    idents = loggers if detailed else [ident for ident in loggers]
    yaml = {"identities": idents}
    dump_as_yaml(yaml)


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-kore
#
# Copyright
#  2019 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""


def find_matching_files(find_filename, location=None):
    """Find all matching files (recursively) from the current working
    directory (default), or from the location specified.

    Args:
        find_filename (str): file name to look for (can be a mask)
        location (str): the location to search for matching files (None
            for cwd)

    Returns:
        list: of matching files

    """
    from os.path import join
    from glob import iglob
    from os import getcwd

    def match_files():
        pathname = join(location, "**", find_filename)
        result = []
        for filename in iglob(pathname, recursive=True):
            result.append(filename)
        return result

    location = location if location else getcwd()
    return match_files()


def put(msg):
    """Put a message to the stdout.

    Args:
        msg (str): the message to write

    """
    from sys import stdout

    stdout.write(msg)


def unlink_if(fpath, verbose=True):
    """Remove the specified file conditionally (exists).

    Args:
        fpath (str): the path to the file
        verbose (bool): If set to false, pipes down the chatter

    """
    from os.path import exists
    from os import unlink

    if exists(fpath):
        if verbose:
            put("Removing file %s\n" % fpath)
        unlink(fpath)


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-kore
#
# Copyright
#  2019 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""
import logging as lgg


def read_input(**kwargs):
    """ Reads input from, either, the command line (nargs returned by
    configargparse) or directly from stdin. The input is parsed for URIs
    -- and if requested, made sure that they are of proper standard.

    Defaults to stdin (if argv is empty or undefined).

    Ignores the comments, for example:

        # A comment line.
        https://foo     # Also ignored.

    Args:
        **kwargs: arbitrary keyword args

    Supported arbitrary keyword args (kwargs):

        nargs (list): As returned by ConfigArgParse, if none is given,
               the input is read from the stdin, instead

        validate_uri (bool): If True, validates that each item is an URI
            Default is False.

        rebuild_uri (bool): If True (default), each item is rebuilt as an URI.
            Useful if you want to make sure that the URIs look
            representable, fixes various URI string presentation issues
            (e.g. 'HtTpS' -> 'https').

            Ignored, unless used together with validate_uri. Default is
            True.

        components_only (bool): If True, each item is returned as a
            list of broken down URI components.

            Also, when True, forces validate_uri (True) and rebuild_uri
            (False). Default is False.

        unique_items_only (bool): If True (default), returns only the
            unique items. Default is True.

    Raises:
        ValueError if validation is enabled and fails (e.g. the input
        contains entries that do not appear to be HTTP URIs)

    Returns:
        list: Parsed values

    """
    unique_items_only = kwargs.get("unique_items_only", True)
    components_only = kwargs.get("components_only", False)
    validate_uri = kwargs.get("validate_uri", False)
    rebuild_uri = kwargs.get("rebuild_uri", True)
    nargs = kwargs.get("nargs")

    rebuild_uri = False if components_only else rebuild_uri
    validate_uri = True if components_only else validate_uri

    if validate_uri:
        # lgg.debug('validate input URIs')
        from urllib.parse import urlparse, urlunparse

    def parse():
        """Parse input for URIs."""

        def add(value):
            """Append a new value to the results."""

            # lgg.debug('input item <%s>' % value)

            if validate_uri:

                uri_components = urlparse(value)
                sch = uri_components.scheme

                if not sch.startswith("http"):
                    raise ValueError("%s: not a valid URI" % value)

                if rebuild_uri:
                    value = urlunparse(uri_components)

                if components_only:
                    value = uri_components

            result.append(value)

        result = []
        if nargs:
            # lgg.debug('parse from nargs')
            for narg in nargs:
                add(narg)
        else:
            # lgg.debug('parse from stdin')
            def read_stdin():
                """Read from stdin."""
                from sys import stdin

                while True:
                    line = stdin.readline().strip()
                    if not line:
                        break
                    line = line.split("#", 1)[0].strip()
                    if line:
                        add(line)

            read_stdin()
        return result

    result = parse()
    if unique_items_only:
        # lgg.debug('return unique items only')
        def unique_items(seq):  # https://stackoverflow.com/a/480227
            """Return unique items in the results, only."""
            seen = set()
            seen_add = seen.add
            return [s for s in seq if not (s in seen or seen_add(s))]

        return unique_items(result)
    return result


if __name__ == "__main__":

    def parse_opts():
        """Parse options."""

        from configargparse import get_parser

        parser = get_parser(add_config_file_help=False)

        parser.add(
            "-b",
            "--verbose",
            help="Enable DEBUG logging level",
            action="store_true",
        )

        parser.add(
            "-v",
            "--validate-uri",
            help="Validate each item as an URI",
            action="store_true",
        )

        parser.add(
            "-u",
            "--unique-items-only",
            help="Return unique items only",
            action="store_true",
        )

        parser.add(
            "-c",
            "--components-only",
            help="Return items as broken down components",
            action="store_true",
        )

        parser.add(
            "-r",
            "--rebuild_uri",
            help="Rebuild items as URIs",
            action="store_true",
        )

        parser.add("uri", nargs="*")
        return parser.parse()

    opts = parse_opts()

    def enable_debug():
        """Enable verbose logging."""
        level = lgg.DEBUG if opts.verbose else lgg.INFO
        lgg.basicConfig(
            level=level, format="[%(levelname)s] %(message)s",
        )

    enable_debug()

    try:
        items = read_input(
            unique_items_only=opts.unique_items_only,
            components_only=opts.components_only,
            validate_uri=opts.validate_uri,
            rebuild_uri=opts.rebuild_uri,
            nargs=opts.uri,
        )
        for item in items:
            lgg.info(item)
        lgg.info("total items %d", len(items))
    except ValueError as error:
        lgg.error(error)

# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-kore
#
# Copyright
#  2019 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""
from signal import getsignal
from signal import signal
from signal import SIGTERM


class GracefulExit:
    """Wraps SIGTERM handling in a with-statement compatible use."""

    __slots__ = ["orig_sigterm", "logger"]

    def __init__(self, logger=None):
        """Initializes the object.

        Args:
            logger (obj): the logger function used to save new entries
                            Can also be None, if logging is not needed.

        """
        self.orig_sigterm, self.logger = (None, logger)

    def __enter__(self):
        """Traps the SIGTERM."""

        self.log("<sigterm> trap signal handler")
        self.orig_sigterm = getsignal(SIGTERM)

        def sigterm_handler(signo, frame):
            """Our SIGTERM handler.

            For lack of a better solution, it currently raises SIGINT
            which we can handle cleanly.

            """
            self.log("<sigterm> signo=%d, frame=%s" % (signo, frame))
            raise KeyboardInterrupt

        signal(SIGTERM, sigterm_handler)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Restore the the SIGTERM handler back to its original state.

        """
        self.log("<sigterm> restore system handler")
        signal(SIGTERM, self.orig_sigterm)

    def log(self, text):
        """Save a new entry to the logger.

        Args:
            text (string): Text to be saved

        """
        if self.logger:
            self.logger(text)


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-kore
#
# Copyright
#  2019 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""


def check_if_positive_integer(argval):
    """Check that the given integer is a positive value."""
    from argparse import ArgumentTypeError

    try:
        value = int(argval)
    except ValueError:
        raise ArgumentTypeError(
            "invalid value type: %s (expected integer)" % argval,
        )
    if value < 1:
        raise ArgumentTypeError("the value must be a positive integer")
    return value


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-yomiel
#
# Copyright
#  2019-2020 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""


def main():
    """main"""
    # from sys import path

    # [print(p) for p in path]

    from yomiel.app import App
    from yomiel import __version__

    pkg_name = "yomiel"
    data_dir = f"{pkg_name}.data"

    App(
        package_name=pkg_name,
        package_data_dir=data_dir,
        config_module=f"{pkg_name}.cache",
        version=__version__,
    ).run()


if __name__ == "__main__":
    main()

# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-yomiel
#
# Copyright
#  2019 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""


def init():
    """Initiates the logging subsystem."""
    from yomiel.cache import logger_paths, opts
    from yomiel.kore.log import log_init

    (logger_file, logger_idents) = log_init(logger_paths)

    from yomiel.log import lg

    lg().debug(
        "subsys/log: configuration file loaded from '%s'", logger_file,
    )

    if opts.logger_idents:
        from yomiel.kore.app import dump_logger_identities

        dump_logger_identities(
            logger_idents, opts.logger_idents_verbose,
        )

    lg().info("log subsystem initiated")


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-comm
#
# Copyright
#  2019 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""


def setup(socket, addr, ssh):
    """Tunnel zmq connection with SSH.

    Args:
        socket (obj): the zmq socket to use
        addr (string): connect to the address via ssh tunnel
        ssh (obj): the namedtuple containing the ssh options

    """
    from zmq.ssh import tunnel_connection

    tunnel_connection(
        socket,
        addr,
        ssh.server,
        keyfile=ssh.key_file,
        password=ssh.password,
        timeout=ssh.timeout,
        paramiko=ssh.paramiko,
    )


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-yomiel
#
# Copyright
#  2019 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""


def init():
    """Initiates the application subsystems."""

    from yomiel.subsys import log

    log.init()

    from yomiel.log import lg

    lg().info("all subsystems initiated")


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-kore
#
# Copyright
#  2019 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""


def log_init(config_paths):
    """Initializes the logger.

    Reads the configuration from a (YAML) file.

    Args:
        config_paths: the list of configuration files (YAML) to read
            Stops at the first existing file.

    Returns:
        str, dict (tuple): the path to the read config file, the
            dictionary holding the identities parsed from the YAML file

    """
    from os.path import expanduser
    import logging.config as lc
    from glob import glob
    from ruamel.yaml import YAML

    # Try to find one of the specified logger config files and load it.
    for path in map(expanduser, config_paths):
        for found in glob(path):
            with open(found) as handle:
                yaml = YAML(typ="safe")
                data = yaml.load(handle)
                lc.dictConfig(data)
                return (found, data["loggers"])

    def set_fallback():
        """Set the fallback logger (if none the config files were found)."""
        from logging import basicConfig, INFO, info

        basicConfig(level=INFO, format="[%(levelname)s] %(message)s")

        info("unable to find %s", config_paths)
        info("using basic logging configuration instead")

        return ("<stdlib:basic>", [])

    return set_fallback()


if __name__ == "__main__":

    def parse_opts():
        """Parse options."""

        from configargparse import get_parser

        parser = get_parser(add_config_file_help=False)

        parser.add(
            "-f",
            "--logger-config",
            help="Logger configuration file to read",
            metavar="FILE",
            required=True,
        )

        parser.add(
            "-l",
            "--logger-ident",
            help="Specify the logger identity to use",
            choices=[
                "default",
                "debug",
                "stderr_file",
                "syslog",
                "file",
                "timed_file",
            ],
            metavar="IDENT",
            default="default",
        )

        return parser.parse()

    opts = parse_opts()
    log_init([opts.logger_config])

    def lg():
        """Return the logger."""
        import logging as lg

        return lg.getLogger(opts.logger_ident)

    lg().debug("Debug level message")
    lg().info("Info level message")

# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-kore
#
# Copyright
#  2019-2020 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""
from ujson import dumps


def json_pprint_formatter(json_object, indent=2, ensure_ascii=False):
    """Return the pretty representation of the given json object.

    This is a convenience "wrapper" function for printing json objects
    in a human-readable representation.

    Args:
        json_object (obj): the json object to use

        indent (int): the used indentation

        ensure_ascii (bool): if True, limits output to ASCII and escapes
            all extended characters above 127 (https://git.io/Jfh2q)

    Returns:
        the human-readable representation of the json object

    """
    return dumps(
        json_object,
        ensure_ascii=ensure_ascii,
        sort_keys=True,
        indent=indent,
    )


def json_pprint(json_object):
    """Pretty print the given json object to the stdout.

    Args:
        json_object: the json object to use

    Returns:

    """
    json_str = json_pprint_formatter(json_object)
    print(json_str)


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-comm
#
# Copyright
#  2019-2020 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""


class InquireError(Exception):
    """Failed to communicate with jomiel."""

    __slots__ = []


def connect(addr, no_linger=True, auth=None, logger=None):
    """Creates the connection socket and connects to jomiel.

    Args:
        addr (string): The endpoint address to connect to

        no_linger (bool): If True, sets the socket option LINGER to 0
            (disable). The default is True.

        auth (obj): The authentication options (if any). Use the
            curve_opts_new, ssh_opts_new and auth_opts_new functions from
            .auth

        logger (obj): The logger to use (if any)

    Returns:
        obj: the created socket

    TODO:
        - Depending on the --auth-mode value, apply whatever
            needs to be applied to establish the connection

    """
    from zmq import Context, REQ, LINGER

    ctx = Context.instance()
    sck = ctx.socket(REQ)
    skip_connect = False

    if auth:
        if auth.curve:
            # Setup CURVE
            from .auth.curve import setup

            setup(sck, auth.curve, logger=logger)
        elif auth.ssh:
            # Setup SSH
            from .auth.ssh import setup

            setup(sck, addr, auth.ssh)
            skip_connect = True

    if not skip_connect:
        sck.connect(addr)

    if no_linger:
        sck.setsockopt(LINGER, 0)

    return sck


def inquire(socket, input_uri, timeout=60):
    """Sends a new metadata inquiry message and awaits for a
    response.

    Exits if the connection attempt timeouts.

    Args:
        socket (obj): the connection socket to us

        input_uri (string): the input URI to inquire the metadata for

        timeout (int): Maximum time in seconds that the program should
            allow the connection to the service to take. Default is 60.

    Raises:
        InquireError if jomiel returned an error

    """

    def inquiry_new():
        """Create a new media inquiry message."""
        from jomiel_messages.protobuf.v1alpha1.message_pb2 import (
            Inquiry,
        )

        inquiry = Inquiry()
        inquiry.media.input_uri = input_uri
        return Inquiry.SerializeToString(inquiry)

    def receive_response():
        """Receive a response message from jomiel.

        Returns:
            obj: The response message

        """
        from jomiel_messages.protobuf.v1alpha1.message_pb2 import (
            Response,
        )
        from jomiel_messages.protobuf.v1alpha1.status_pb2 import (
            STATUS_CODE_OK,
        )

        data = socket.recv()
        resp = Response()
        resp.ParseFromString(data)

        if resp.status.code != STATUS_CODE_OK:
            raise InquireError(
                "%s (status=%d, error=%d, http=%d)"
                % (
                    resp.status.message,
                    resp.status.code,
                    resp.status.error,
                    resp.status.http.code,
                ),
            )

        return resp

    inquiry = inquiry_new()
    socket.send(inquiry)

    from zmq import Poller, POLLIN

    poller = Poller()
    poller.register(socket, POLLIN)

    connect_timeout = int(timeout) * 1000  # to msec

    if poller.poll(connect_timeout):
        return receive_response()
    raise OSError("connection timed out")


def to_json(message, minified=False, stream=None, **kwargs):
    """Returns the JSON of the given protobuf message.

    Args:
        Message (obj): The protobuf message to be converted

        minified (bool): If True, the resulting JSON will be minified,
            otherwise (default) a human-readable representation is
            returned.

        stream (obj): Unless None, write the json to stream

        **kwargs (list): arbitrary keyword args (to be passed as such to
            ujson.dumps)

    Supported arbitrary keyword args (kwargs):
        See the `ujson` documentation at <https://git.io/Jfhr7>.

    Returns:
        str: the resulting JSON -- unless stream is None, in which case
            the resulting JSON is written to the stream and nothing is
            returned

    Notes:
        - google.protobuf.json_format.MessageToJson returns a
          human-readable representation of the data

        - For a minified representation, use the ujson module

    """
    from google.protobuf.json_format import MessageToJson

    rval = MessageToJson(message)

    if minified:
        from ujson import loads, dumps

        loaded = loads(rval)
        rval = dumps(loaded, **kwargs)

    if stream:
        stream.write(rval)
    else:
        return rval


def to_yaml(message, stream=None):
    """Returns the YAML of the given protobuf message.

    Args:
        Message (obj): The protobuf message to be converted

    Returns:
        str: the resulting YAML

    """
    from google.protobuf.json_format import MessageToDict

    data = MessageToDict(message)

    from ruamel.yaml import YAML, round_trip_dump

    yaml = YAML(typ="safe")
    yaml.default_flow_style = False

    stream.write("---\n")
    return round_trip_dump(data, stream)


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-kore
#
# Copyright
#  2019-2020 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""

try:  # py38+
    from importlib.metadata import version as metadata_version
    from importlib.metadata import PackageNotFoundError
except ModuleNotFoundError:
    from importlib_metadata import version as metadata_version
    from importlib_metadata import PackageNotFoundError


def package_version(package_name, destination):
    """Returns the package version string

    Args:
        package_name (str): the package name to look up
        destination (list): the list to store the result (tuple) to

    """
    try:
        version = metadata_version(package_name)
    except PackageNotFoundError:
        version = "<unavailable>"

    if package_name == "pyzmq":
        from zmq import zmq_version

        version = "{} (libzmq version {})".format(
            version, zmq_version(),
        )

    destination.append((package_name, version))


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-yomiel
#
# Copyright
#  2019 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""


def lg():
    """Returns the logger instance used to print to the logging
    subsystem to record new events.

    The subsystem is configured via a separate logger YAML configuration
    file. The configuration supports different logger identities.

    To use this function (lg):

        from jomiel import lg
        lg().debug('foo=%s' % foo)

    Returns
        The logger instance

    """
    from yomiel.cache import opts
    import logging as lg

    return lg.getLogger(opts.logger_ident)


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-comm
#
# Copyright
#  2019 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""
from collections import namedtuple


def curve_opts_new(server_public_key_file, client_key_file):
    """Returns a namedtuple containing the curve options."""

    curve_opts = namedtuple(
        "curve_opts", "server_public_key_file, client_key_file",
    )

    return curve_opts(server_public_key_file, client_key_file)


def ssh_opts_new(
    ssh_server, ssh_key_file, ssh_password, ssh_timeout, ssh_paramiko,
):
    """Return a namedtuple containing the ssh options."""

    ssh_opts = namedtuple(
        "ssh_opts", "server, key_file, password, timeout, paramiko",
    )

    return ssh_opts(
        ssh_server,
        ssh_key_file,
        ssh_password,
        ssh_timeout,
        ssh_paramiko,
    )


def auth_opts_new(curve=None, ssh=None):
    """Return a namedtuple containing either curve or ssh options.

    You can use the `curve_opts_new` and `ssh_opts_new` functions with this
    function.

    Args:
        curve (obj): the namedtuple created with `curve_opts_new`
        ssh (obj): the namedtuple created with `ssh_opts_new`

    Notes:
        - You must set eithe curve(:arg:) or ssh(:arg:), not both or
          neither

    Raises:
        ValueError if any of the args are invalid

    Returns:
        obj: New namedtuple containing either curve or ssh options

    """
    if not curve and not ssh:
        raise ValueError("curve and ssh cannot both be None")

    elif curve and ssh:
        raise ValueError(
            "you can set only either curve or ssh, not both",
        )

    auth_opts = namedtuple("auth_opts", "curve, ssh")
    return auth_opts(curve, ssh)


# vim: set ts=4 sw=4 tw=72 expandtab:



#
# jomiel-yomiel
#
# Copyright
#  2019 Toni Gündoğdu
#
#
# SPDX-License-Identifier: Apache-2.0
#
"""TODO."""
from yomiel.kore.app import App as KoreApp
from yomiel.kore.app import exit_normal
from yomiel.kore.opts import check_if_positive_integer
from yomiel.log import lg


class App(KoreApp):
    """Implements the application."""

    __slots__ = []

    def version_long_packages(self):
        """Return the required package (list) for --version-long."""
        return [
            "jomiel_messages",
            "configargparse",
            "ruamel.yaml",
            "protobuf",
            "pyzmq",
            "ujson",
        ]

    def run(self):
        """Application entry point; executes the app."""
        from configargparse import get_parser

        parser = get_parser()

        parser.add(
            "uri", metavar="<uri>", nargs="*", help="the URIs to parse",
        )

        parser.add(
            "-o",
            "--output-format",
            help="Print messages in the specified data serialization format",
            choices=["raw", "json", "yaml", "terse"],
            metavar="[raw|json|yaml|terse]",
            default="raw",
        )

        def jomiel_group():
            """Add the jomiel options group."""
            grp = parser.add_argument_group("jomiel")

            grp.add(
                "-r",
                "--router-endpoint",
                help="jomiel router endpoint address to connect to",
                default="tcp://localhost:5514",
                metavar="ADDR",
            )

            grp.add(
                "-t",
                "--connect-timeout",
                help="""Maximum time in seconds that the program should
                        allow the connection to the service to take""",
                type=check_if_positive_integer,
                default=60,
                metavar="TIME",
            )

        jomiel_group()

        def debug_group():
            """Add the debug options group."""
            grp = parser.add_argument_group("debug")

            grp.add(
                "-m",
                "--debug-minify-json",
                help="Minify JSON messages in the logger",
                action="store_true",
            )

        debug_group()

        def auth_group():
            """Add the auth option group."""
            grp = parser.add_argument_group("auth")

            grp.add(
                "--auth-mode",
                help="Select authentication mode",
                choices=["none", "curve", "ssh"],
                metavar="[none|curve|ssh]",
                default="none",
            )

        auth_group()

        def curve_group():
            """Add the auth-curve option group."""
            grp = parser.add_argument_group("auth: curve")

            grp.add(
                "--curve-server-public-key-file",
                help="""Public CURVE certificate key file to use for
                        connecting to jomiel""",
                default=".curve/server.key",
                metavar="FILE",
            )

            grp.add(
                "--curve-client-key-file",
                help="""Secret client CURVE key file to use for connecting
                        to jomiel""",
                default=".curve/client.key_secret",
                metavar="FILE",
            )

        curve_group()

        def ssh_group():
            """Add the auth-ssh option group."""
            grp = parser.add_argument_group("auth: ssh")

            grp.add(
                "--ssh-server",
                help="SSH server to connect to",
                metavar="user@server:port",
            )

            grp.add(
                "--ssh-key-file",
                help="Path to the key file to use",
                metavar="FILE",
            )

            grp.add(
                "--ssh-password",
                help="Password to the SSH server",
                metavar="PASSWD",
            )

            grp.add(
                "--ssh-timeout",
                help="""Time (in seconds) after which no activity will
                        result in the tunnel closing""",
                default=60,
                type=check_if_positive_integer,
                metavar="TIME",
            )

            grp.add(
                "--ssh-paramiko",
                help="Use paramiko instead of pexpect",
                action="store_true",
            )

        ssh_group()

        def dump_terse_response(stdout, media_response):
            """Dump a terse response

            Args:
                stdout (obj): the std output stream to dump the object to
                media_response (obj): the media response object to dump

            """
            stdout.write("---\ntitle: " + media_response.title + "\n")
            stdout.write("quality:\n")

            def get_terse_quality_string():
                """Return terse string for a stream quality."""
                return "  profile: {}\n    width: {}\n    height: {}\n".format(
                    stream_quality.profile,
                    stream_quality.width,
                    stream_quality.height,
                )

            for stream in media_response.stream:
                stream_quality = stream.quality
                quality_string = get_terse_quality_string()
                stdout.write(quality_string)

        def dump_metadata(response):
            """Print the metadata to standard output."""

            def has_stream():
                """Has stream data."""
                return response.media.stream

            def has_image():
                """Has image data."""
                return response.media.image

            if has_stream():
                lg().info("has video")
            elif has_image():
                lg().info("has image")
            else:
                handle_error("unexpected response (empty media lists)")

            from sys import stdout

            if "raw" in opts.output_format:
                stdout.write(str(response.media))
            else:
                if "json" in opts.output_format:
                    from yomiel.comm import to_json

                    to_json(
                        response.media,
                        minified=opts.debug_minify_json,
                        stream=stdout,
                    )
                elif "yaml" in opts.output_format:
                    from yomiel.comm import to_yaml

                    to_yaml(response.media, stream=stdout)
                elif "terse" in opts.output_format:
                    dump_terse_response(stdout, response.media)
                else:
                    handle_error(
                        "unexpected --output-format value (%s)"
                        % opts.output_format,
                    )

        def determine_auth_opts():
            """Determine whether auth should be used."""
            from yomiel.comm.auth import auth_opts_new

            rval = None

            if "curve" in opts.auth_mode:
                from yomiel.comm.auth import curve_opts_new

                curve_opts = curve_opts_new(
                    opts.curve_server_public_key_file,
                    opts.curve_client_key_file,
                )

                rval = auth_opts_new(curve=curve_opts)

            elif "ssh" in opts.auth_mode:

                if not opts.ssh_server:
                    handle_error(
                        "argument --ssh-server: conflicting option "
                        "string (None) when used with --auth-mode=ssh",
                    )

                from yomiel.comm.auth import ssh_opts_new

                ssh_opts = ssh_opts_new(
                    opts.ssh_server,
                    opts.ssh_key_file,
                    opts.ssh_password,
                    opts.ssh_timeout,
                    opts.ssh_paramiko,
                )

                rval = auth_opts_new(ssh=ssh_opts)

            return rval

        def main_loop():
            """This is the main loop for the application."""
            input_uri = read_input(opts.uri)
            auth_opts = determine_auth_opts()

            from yomiel.comm import connect, inquire, InquireError

            try:
                lg().info(
                    "connect to %s (timeout=%d)",
                    opts.router_endpoint,
                    opts.connect_timeout,
                )

                sck = connect(
                    opts.router_endpoint, auth=auth_opts, logger=lg(),
                )

                for uri in input_uri:
                    lg().info("inquire <%s>", uri)

                    resp = inquire(
                        sck, uri, timeout=opts.connect_timeout,
                    )

                    dump_metadata(resp)

            except KeyboardInterrupt:
                sigint()

            except OSError as msg:
                handle_error(msg)

            except InquireError as msg:
                handle_error(msg)

            exit_normal()

        opts = super().parse_opts(parser)
        lg().disabled = "terse" in opts.output_format

        from yomiel.subsys import init

        init()
        main_loop()


def handle_error(msg):
    """Handle an error."""
    lg().disabled = False
    lg().error(msg)
    from yomiel.kore.app import exit_error

    exit_error()


def read_input(nargs):
    """Read input (URIs) from the user.

    Returns:
        list: containing the parsed input URIs

    """
    from yomiel.kore.input import read_input as parse

    try:
        input_uri = parse(validate_uri=True, nargs=nargs)
        if not input_uri:
            handle_error("an input URI was not given")
    except KeyboardInterrupt:
        sigint()
    except ValueError as msg:
        handle_error(msg)
    return input_uri


def sigint():
    """Handle SIGINT."""
    lg().error("signal interrupt")
    exit_normal()


# vim: set ts=4 sw=4 tw=72 expandtab:
